
#-------------------------------------------------
# Small matrix multiplication untility for OpenGL
# Matrices are stored in column major order
#-------------------------------------------------

import numpy as np
import math

def identity_matrix(n=4):
    return np.identity(n)

def scale_matrix(scale):
    return np.diag([scale, scale, scale, 1.0])

def translation_matrix(direction):
    return np.array([[1.0, 0.0, 0.0, 0.0],
                     [0.0, 1.0, 0.0, 0.0],
                     [0.0, 0.0, 1.0, 0.0],
                     [direction[0], direction[1], direction[2], 1.0]])

# https://en.wikipedia.org/wiki/Rotation_matrix#Axis_and_angle
def rotation_matrix(angle, axis):
    s = math.sin(angle)
    c = math.cos(angle)
    C = 1 - c

    norm = np.linalg.norm(axis[:3])
    if norm == 0:
        return np.identity(3)
    
    axis /= np.linalg.norm(axis[:3])
    x = axis[0]
    y = axis[1]
    z = axis[2]
    M = np.array([[x*x*C + c,   y*x*C + z*s, z*x*C-y*s],
                  [x*y*C - z*s, y*y*C + c,   z*y*C+x*s],
                  [x*z*C + y*s, y*z*C - x*s, z*z*C + c]])
    return M

def normalize(vector):
    vector = np.array(vector, dtype=np.float64, copy=True)
    vector /= np.linalg.norm(vector)
    return vector

# matrix for camera at position 'eye' looking at position 'center'
def look_at_matrix(eye, center, up):
    z = normalize(eye - center)
    x = normalize(np.cross(up, z))
    y = np.cross(z, x)

    # pos = np.add(center, eye)

    M = np.array([[ x[0],            y[0],            z[0],           0.0],
                  [ x[1],            y[1],            z[1],           0.0],
                  [ x[2],            y[2],            z[2],           0.0],
                  [-np.dot(x, eye), -np.dot(y, eye), -np.dot(z, eye), 1.0]])
    return M

# returns the arcball vector given the position of the mouse and dimensions of the screen
# https://en.wikibooks.org/wiki/OpenGL_Programming/Modern_OpenGL_Tutorial_Arcball
def arcball_vector(x, y, width, height):
    p = np.array([2.0 * x / width - 1.0, 1.0 - 2.0 * y / height, 0])
    mag_squared = p[0]**2 + p[1]**2

    if mag_squared <= 1.0:
        p[2] = math.sqrt(1.0 - mag_squared)
    else:
        p /= math.sqrt(mag_squared)
    return p

def sphere_rotation_matrix(position, up, m0, m1, width, height):
    radius = np.linalg.norm(position)

    z = normalize(position)
    x = normalize(np.cross(up, z))
    y = np.cross(z, x)

    M = np.array([[x[0], y[0], z[0]],
                  [x[1], y[1], z[1]],
                  [x[2], y[2], z[2]]])

    if m0 == m1:
        return identity_matrix(3)

    p0 = arcball_vector(m0[0], m0[1], width, height)
    p1 = arcball_vector(m1[0], m1[1], width, height)
    angle = math.acos(np.dot(p0, p1))
    axis = np.cross(p0, p1)
    axis = np.matmul(M, axis)
    return rotation_matrix(angle, axis)

def sphere_translation(position, up, direction):
    z = normalize(position)
    x = normalize(np.cross(up, z))
    y = np.cross(z, x)

    M = np.array([[x[0], y[0], z[0]],
                  [x[1], y[1], z[1]],
                  [x[2], y[2], z[2]]])
    #M = np.linalg.inv(M)
    direction = np.matmul(M, direction)
    return direction

# fovy should be in degrees
# https://unspecified.wordpress.com/2012/06/21/calculating-the-gluperspective-matrix-and-other-opengl-matrix-maths/
def perspective_matrix(fovy, aspectRatio, near, far):
    f = 1.0 / math.tan(fovy * math.pi / 360.0)
    return np.array([[f / aspectRatio, 0, 0, 0], [0, f, 0, 0], [0, 0, (far + near)/(near - far), -1], [0, 0, 2*far*near/(near-far), 0]])

def arcball_matrix(m0, m1, width, height, speedMultiplier = 1.0, viewInverse = None):
    p0 = arcball_vector(m0[0], m0[1], width, height)
    p1 = arcball_vector(m1[0], m1[1], width, height)

    # p0 and p1 should already be normailized
    angle = speedMultiplier * math.acos(np.dot(p0, p1))
    axis = np.cross(p0, p1)

    if not viewInverse is None:
        axis = np.append(axis, 0.0)
        axis = np.matmul(axis, viewInverse)

    M = rotation_matrix(angle, axis)
    return M
